import json

from django.http import HttpResponse

from .constants import ERROR
from .prometheus import (
    PROMETHEUS_AVAILABLE,
    PROMETHEUS_SCHEMA,
    prometheus_response,
)


def should_output_prometheus(request):
    """
    Decide if Canary should output for Prometheus based on the request.

    The Python Prometheus client only supports the plain text format and not
    the protocol buffer format.

    Prometheus formats:
    http://prometheus.io/docs/instrumenting/exposition_formats/
    """
    if not PROMETHEUS_AVAILABLE:
        return False

    if request.GET.get('format') == 'prometheus':
        return True

    return PROMETHEUS_SCHEMA in request.META.get('HTTP_ACCEPT', '')


def json_response(canary, indent=2):
    """
    Return a 200 if the database is available, or a 503 otherwise.
    """
    result = canary.check()
    content = json.dumps(result, indent=indent)
    status = 503 if result['status'] == ERROR else 200
    return HttpResponse(
        content,
        status=status,
        content_type='application/json'
    )


def status(request, canary, **kwargs):
    if should_output_prometheus(request):
        return prometheus_response(canary, **kwargs)
    else:
        return json_response(canary, **kwargs)



from __future__ import unicode_literals

import re

from django.core.exceptions import ImproperlyConfigured
import requests

from ..constants import OK, WARNING, ERROR
from ..decorators import timed
from . import Resource


STATUS_MAPPING = {
    'green': OK,
    'yellow': WARNING,
    'red': ERROR,
}


class ElasticSearch(Resource):
    """
    Checks the status of an ElasticSearch cluster.

    ElasticSearch has its own health status we can wrap. See:
    http://www.elastic.co/guide/en/elasticsearch/reference/1.x/cluster-health.html

    :param name: The name of this canary.
    :param host: The ElasticSearch host to connect to.
    :param timeout: The maximum time (in seconds) to wait for a response.
    """
    def __init__(self, name, host, timeout=30.0, **kwargs):
        self.host = self._normalize_host(host)
        self.timeout = timeout
        kwargs.setdefault('description', self.host)
        super(ElasticSearch, self).__init__(name, **kwargs)

    @timed(threshold=0.1)
    def check(self):
        result = super(ElasticSearch, self).check()
        try:
            health_url = '{}/_cluster/health'.format(self.host)
            response = requests.get(health_url, timeout=self.timeout)
        except requests.exceptions.RequestException as e:
            return dict(result, status=ERROR, error=str(e))
        data = response.json()
        # WISHLIST there's more data that might be worth passing through
        # besides just the status, see the link in the docstring
        status = STATUS_MAPPING[data['status']]
        return dict(result, status=status)

    def _normalize_host(self, host):
        """
        Accept multiple ways to specify the ElasticSearch host.

        This way, projects can store the connection string in the format most
        convenient in the project.
        """
        # urlparse won't work here because it interprets the host as the path
        matches = re.match('(?P<scheme>https?:)?(//)?(?P<host>[^:]+)'
                           '(:(?P<port>\d+))?.*', host)
        if not matches:
            raise ImproperlyConfigured('{} does not look like an ElasticSearch'
                                       ' connection string'.format(host))
        data = matches.groupdict()
        return '{}//{}:{}'.format(
            data['scheme'] or 'http:',
            data['host'],
            data['port'] or '9200',
        )



from __future__ import absolute_import

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django_rq import queues
from redis import RedisError
from rq import Worker

from ..constants import OK, WARNING, ERROR
from ..decorators import timed
from . import Resource


class DjangoRQ(Resource):
    """
    Checks the status of any RQ queues configured by Django.

    All queues must share the same Redis connection.
    """

    def __init__(
            self,
            name='rq',
            description='Django RQ',
            queues=None,
            **kwargs):
        self.queue_names = queues or settings.RQ_QUEUES.keys()
        super(DjangoRQ, self).__init__(
            name=name,
            description=description,
            **kwargs
        )

    @timed
    def check(self):
        result = super(DjangoRQ, self).check()
        status = result.get('status', OK)
        try:
            result['queues'] = {}
            for queue in self.get_queues():
                queue_result = self.check_queue(queue)
                status = self.get_status_for_queue_result(status, queue_result)
                result['queues'][queue.name] = queue_result

            return dict(result, status=status)
        except (ImproperlyConfigured, RedisError) as e:
            return dict(result, status=ERROR, error=str(e))

    def get_queues(self):
        return queues.get_queues(*self.queue_names)

    def check_queue(self, queue):
        n_jobs = queue.count
        n_workers = self.count_queue_workers(queue)
        url = self.get_queue_url(queue)
        return {'n_jobs': n_jobs, 'n_workers': n_workers, 'url': url}

    def get_queue_url(self, queue):
        configuration = queue.connection.connection_pool.connection_kwargs
        return 'redis://{host}:{port}/{db}/'.format(**configuration)

    def count_queue_workers(self, queue):
        connection = queues.get_connection(queue.name)
        workers = Worker.all(connection=connection)
        return len([_ for _ in workers if queue in _.queues])

    def get_status_for_queue_result(self, status, result):
        if status == OK and result['n_jobs'] > result['n_workers']:
            return WARNING
        else:
            return status



from django.conf import settings
from django.db import connections, DatabaseError

from ..constants import OK, ERROR
from ..decorators import timed
from . import Resource

try:
    # Django 1.7 and later should wrap this exception in a DatabaseError
    from MySQLdb import OperationalError as MySQLError
except ImportError:
    MySQLError = DatabaseError


class Database(Resource):
    """
    Checks the status of a database.

    :param connection: A database connection to query for availability.
    :param statements: The sequence of SQL statements to execute.
    """
    default_statement = 'SELECT 1'

    def __init__(self, name, connection, statements=None, **kwargs):
        self.connection = connection
        self.statements = statements or [self.default_statement]
        super(Database, self).__init__(name, **kwargs)

    @timed(threshold=0.1)
    def check(self):
        result = super(Database, self).check()
        try:
            cursor = self.connection.cursor()
            map(cursor.execute, self.statements)
            return dict(result, status=OK)
        except (DatabaseError, MySQLError) as e:
            return dict(result, status=ERROR, error=str(e))


class DjangoDatabase(Database):
    """
    Checks a database resource with a Django database connection.
    """
    def __init__(self, name='database', database='default', **kwargs):
        super(DjangoDatabase, self).__init__(
            name=name,
            description=self.get_description(database),
            connection=connections[database],
            **kwargs
        )

    def get_description(self, database):
        template = 'Database "{database}" at {ENGINE}://{HOST}/{NAME}'
        configuration = settings.DATABASES[database].copy()
        configuration.setdefault('HOST', '')
        return template.format(database=database, **configuration)



from .constants import OK, WARNING, ERROR
from .decorators import timed
from .prometheus import PrometheusCanaryMixin
from .resources.projects import Project, GitProject


class Canary(PrometheusCanaryMixin, object):
    """
    Reports the status of a project and its dependencies.
    """
    project_class = Project

    def __init__(self, name, version=None, root=None, resources=None):
        self.project = self.project_class(name, version=version, root=root)
        self.resources = resources or []
        super(Canary, self).__init__(name, version, root, resources)

    @timed
    def check(self):
        metadata = self.project.check()
        resources = {r.name: r.check() for r in self.resources}
        status = self.get_worst_status(results=resources.values())
        return dict(metadata, status=status, resources=resources)

    def get_worst_status(self, results):
        status = OK
        for result in results:
            resource_status = result.setdefault('status', OK)
            if resource_status == ERROR:
                return ERROR
            elif status == OK and resource_status == WARNING:
                status = WARNING

        return status


class GitCanary(Canary):
    """
    Reports the status and revision of a Git project and dependencies.
    """
    project_class = GitProject



import functools
import time

from .constants import OK, WARNING


def timed(check=None, threshold=1.0):
    """
    Wrap a status check with a latency in seconds.

    :param threshold:
        The number of seconds before the status is elevated to WARNING.
    """
    def wrap_check(check):
        @functools.wraps(check)
        def wrapped_check(*args, **kwargs):
            start = time.time()
            result = check(*args, **kwargs)
            latency = time.time() - start

            status = result['status']
            if status == OK and latency >= threshold:
                status = WARNING

            return dict(result, status=status, latency=latency)

        return wrapped_check

    return wrap_check(check) if check else wrap_check



from django.http import HttpResponse
try:
    from prometheus_client import CONTENT_TYPE_LATEST, Gauge, generate_latest
    PROMETHEUS_AVAILABLE = True
except ImportError:
    PROMETHEUS_AVAILABLE = False


PROMETHEUS_SCHEMA = 'schema="prometheus/telemetry"'


class PrometheusCanaryMixin(object):
    """
    Enable a Canary to give reports that can be consumed by Prometheus.

    The keys of the prometheus dict aren't that meaningful, so they're named
    after how to access the data from the canary dict.
    """
    def __init__(self, name, *args, **kwargs):
        if not PROMETHEUS_AVAILABLE:
            return
        self.prometheus = {
            'latency': Gauge(
                name, 'latency_seconds', ['resource']),
            'queues.n_jobs': Gauge(
                name, 'redis_queue_jobs_count', ['resource', 'queue']),
            'queues.n_workers': Gauge(
                name, 'redis_queue_workers_count', ['resource', 'queue']),
        }


def prometheus_response(canary):
    """
    Return a HTTP Canary response that Prometheus can understand.
    """
    result = canary.check()
    prom = canary.prometheus
    for name, resource in result['resources'].items():
        if 'latency' in resource:
            prom['latency'].labels(name).set(resource['latency'])
        if 'queues' in resource:
            for queue_name, queue_stats in resource['queues'].items():
                (prom['queues.n_jobs']
                    .labels(name, queue_name)
                    .set(queue_stats['n_jobs']))
                (prom['queues.n_workers']
                    .labels(name, queue_name)
                    .set(queue_stats['n_workers']))

    return HttpResponse(
        generate_latest(),
        content_type=CONTENT_TYPE_LATEST,
    )



import os
import sys


# Project
#########

DEBUG = True

ROOT = os.path.abspath(os.path.dirname(__file__))
sys.path.append(ROOT)

VERSION = '1.2.3'

ROOT_URLCONF = 'tests.projects.example'

SECRET_KEY = 'foo'


# Databases
###########

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        'USER': 'bar',
        'PASSWORD': 'baz',
    }
}


# RQ
####

RQ_QUEUES = {
    'default': {
        'URL': 'redis://foo',
        'DEFAULT_TIMEOUT': 360,
        'DB': 0,
    },
    'low': {
        'URL': 'redis://foo',
        'DEFAULT_TIMEOUT': 360,
        'DB': 0,
    },
    'high': {
        'URL': 'redis://foo',
        'DEFAULT_TIMEOUT': 360,
        'DB': 0,
    },
}


# Services
##########

HTTP_ENDPOINT = 'http://service/'

HTTP_ENDPOINT_WITH_CANARY = 'http://service/_status/'


# Canary
########

from canary_endpoint import Canary
from canary_endpoint.resources.databases import DjangoDatabase
from canary_endpoint.resources.rq import DjangoRQ
from canary_endpoint.resources.search import ElasticSearch
from canary_endpoint.resources.services import Service, ServiceWithCanary

canary = Canary('example', root=ROOT, version=VERSION, resources=[
    DjangoDatabase(statements=['SELECT 1 FROM foo;']),
    DjangoRQ(queues=['default']),
    ElasticSearch('es_resource', host='http://elasticsearch.example:9200'),
    Service('foo', url=HTTP_ENDPOINT),
    ServiceWithCanary('bar', url=HTTP_ENDPOINT_WITH_CANARY),
])


# URLs
######

from canary_endpoint.views import status
from django.conf.urls import patterns, url

urlpatterns = patterns('', url(r'^_status/$', status, {'canary': canary}))



from ..constants import OK


class Resource(object):
    """
    Represents a Resource with a status.

    "status" may be one of:

    - "ok", the resource is available
    - "warning", the resource may not be available
    " "error", the resource is unavailable
    """
    def __init__(self, name, description, **kwargs):
        self.name = name
        self.description = description

    def check(self):
        return {'description': self.description, 'status': OK}



from subprocess import CalledProcessError
import json
import re

from mock import patch, Mock
import responses


class MockMixin(object):
    def patch(self):
        pass

    def unpatch(self):
        pass


class MockTestCaseMixin(MockMixin):
    def setUp(self):
        super(MockTestCaseMixin, self).setUp()
        self.patch()

    def tearDown(self):
        self.unpatch()
        super(MockTestCaseMixin, self).tearDown()


class MockTimeMixin(MockMixin):
    def patch(self):
        super(MockTimeMixin, self).patch()
        self.time_patcher = patch('canary_endpoint.decorators.time')
        self.mock_time_module = self.time_patcher.start()
        self.mock_duration()

    def unpatch(self):
        self.time_patcher.stop()
        super(MockTimeMixin, self).unpatch()

    def mock_duration(self, initial=0.0, step=0.0):
        timings = self.get_timings(initial=initial, step=step)
        self.mock_time_module.time.side_effect = timings

    def get_timings(self, initial=0.0, step=0.0):
        # Yield an initial time for the canary
        yield initial
        while True:
            # Yield times for each resource
            yield initial
            yield initial + step
            initial = initial + step


class MockDatabaseMixin(MockMixin):
    database = 'default'

    def patch(self):
        super(MockDatabaseMixin, self).patch()
        execute = 'django.db.backends.util.CursorDebugWrapper.execute'
        self.cursor_patcher = patch(execute, autospec=True)
        self.mock_execute = self.cursor_patcher.start()

    def unpatch(self):
        self.cursor_patcher.stop()
        super(MockDatabaseMixin, self).unpatch()

    def mock_database_error(self, message=None):
        from django.db import DatabaseError
        self.mock_execute.side_effect = DatabaseError(message)


class MockRequestsMixin(object):
    """If you need to mock `requests` GET requests in your Canary tests."""
    elasticsearch_urls = re.compile(r'https?://elasticsearch.*')
    service_urls = re.compile(r'http://service.*')

    def setUp(self):
        super(MockRequestsMixin, self).setUp()
        responses.start()

    def tearDown(self):
        responses.stop()
        responses.reset()
        super(MockRequestsMixin, self).tearDown()

    def reset_responses(self):
        responses.reset()

    def mock_ok_service(self, data=None):
        self.mock_service_response(status_code=200, data=data)

    def mock_error_service(self, data=None):
        self.mock_service_response(status_code=500, data=data)

    def mock_service_response(self, status_code, data=None):
        responses.add(
            responses.GET, self.service_urls,
            body=json.dumps(data) if data else '',
            status=status_code,
            content_type='application/json')

    def mock_ok_elasticsearch(self):
        responses.add(
            responses.GET, self.elasticsearch_urls,
            body='{"status": "green"}',
            content_type='application/json')

    def mock_error_elasticsearch(self, body='{"status": "red"}'):
        responses.add(
            responses.GET, self.elasticsearch_urls,
            body=body,
            content_type='application/json')


class MockRQMixin(MockMixin):

    rq_default_configuration = {'host': 'localhost', 'port': '6379', 'db': 0}

    def patch(self):
        super(MockRQMixin, self).patch()
        self.get_queue_patcher = patch(
            'canary_endpoint.resources.rq.queues.get_queue',
            autospec=True
        )
        self.worker_patcher = patch('canary_endpoint.resources.rq.Worker')
        self.mock_get_queue = self.get_queue_patcher.start()
        self.mock_get_queue.side_effect = self.get_mock_queue
        self.mock_worker = self.worker_patcher.start()
        self.rq_queues = {}

    def unpatch(self):
        self.mock_get_queue = self.get_queue_patcher.stop()
        self.mock_worker = self.worker_patcher.stop()
        super(MockRQMixin, self).unpatch()

    def mock_queue(self, name, configuration=None, count=0):
        mock_queue = Mock()
        mock_queue.name = name
        mock_queue.count = count
        configuration = configuration or self.rq_default_configuration
        mock_queue.connection.connection_pool.connection_kwargs = configuration
        self.rq_queues[name] = mock_queue

    def get_mock_queue(self, name, **kwargs):
        return self.rq_queues[name]

    def mock_redis_error(self, message=None):
        from redis.exceptions import RedisError
        self.mock_get_queue.side_effect = RedisError(message)


class MockGitMixin(MockMixin):
    def patch(self):
        super(MockGitMixin, self).patch()
        self.git_output_patcher = patch(
            'canary_endpoint.resources.projects'
            '.subprocess.check_output'
        )
        self.mock_git_output = self.git_output_patcher.start()

    def unpatch(self):
        self.mock_git_output = self.git_output_patcher.stop()

    def mock_git_revision(self, revision):
        self.mock_git_output.return_value = revision

    def mock_git_error(self, command='foo', code=1, output=None):
        output = output or "'{}' is not a git command.".format(command)
        exception = CalledProcessError(code, command, output)
        self.mock_git_output.side_effect = exception

    def mock_git_not_found(self, message='command not found: git'):
        self.mock_git_output.side_effect = OSError(message)



import subprocess

from ..constants import ERROR
from . import Resource


class GitMixin(object):
    """
    Wraps a project check to include the current Git revision.
    """
    def check(self):
        result = super(GitMixin, self).check()
        try:
            revision = self.get_revision()
            return dict(result, revision=revision)
        except (OSError, subprocess.CalledProcessError) as e:
            return dict(result, revision=None, status=ERROR, error=str(e))

    def get_revision(self):
        command = ('git', 'rev-parse', 'HEAD')
        return subprocess.check_output(command, cwd=self.root).strip()


class Project(Resource):
    """
    Checks the status of a project.
    """
    def __init__(self, name, root, version=None, **kwargs):
        self.name = name
        self.root = root
        self.version = version

    def check(self):
        return {'name': self.name, 'version': self.get_version()}

    def get_version(self):
        return self.version


class GitProject(GitMixin, Project):
    """
    Checks the status of a Git project.
    """
    pass



import os
from setuptools import setup, find_packages


namespace = {}
version_path = os.path.join('canary_endpoint', 'version.py')
exec(open(version_path).read(), namespace)


setup(
    name='django-canary-endpoint',
    version=namespace['__version__'],
    description='Resource health checks for Django.',
    long_description=open('README.md').read(),
    author='TabbedOut',
    author_email='dev@tabbedout.com',
    url='https://github.com/TabbedOut/django-canary-endpoint',
    license='MIT License',
    packages=find_packages(exclude=('tests', 'docs')),
    test_suite='tests',
    install_requires=[
        'django-rq>=0.7.0,<1',
        'requests>=2.0.0,<3',
        'responses<1',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
    ],
)



import requests

from ..constants import OK, WARNING, ERROR
from ..decorators import timed
from . import Resource


class Service(Resource):
    """
    Checks the status of a service with an HTTP GET request.

    :param url: The URL of the service's status endpoint.
    :param headers: A dictionary of headers to include with the check.
    :param timeout: The number of seconds before an error is returned.
    """
    def __init__(self, name, url, headers=None, timeout=1.0, **kwargs):
        self.url = url
        self.timeout = timeout
        self.headers = headers
        super(Service, self).__init__(name=name, description=url, **kwargs)

    @timed
    def check(self):
        result = super(Service, self).check()
        try:
            self.last_response = None
            self.last_response = self.get_http_response()
            self.last_response.raise_for_status()
            return dict(result, status=OK)
        except requests.RequestException as e:
            return dict(result, status=ERROR, error=str(e))

    def get_http_response(self):
        return requests.get(
            self.url,
            headers=self.headers,
            timeout=self.timeout
        )


class ServiceWithCanary(Service):
    """
    Checks the status of a service with a canary endpoint.

    Reports an error if no JSON can be decoded from the remote canary.

    :param url: The URL of the service's canary endpoint.
    """
    def check(self):
        result = super(ServiceWithCanary, self).check()
        status = result.get('status')
        try:
            service_result = self.last_response.json()
            service_status = service_result.get('status')
            if status == OK and service_status == WARNING:
                status = WARNING
            elif service_status == ERROR:
                status = ERROR
            return dict(result, status=status, result=service_result)
        except (AttributeError, ValueError) as e:
            return dict(result, status=ERROR, result=None, error=str(e))
